/*
 * implements alarm handling
 */

#include "libocp_alarm.h"
#include "../src/liboculus.h"
#include <string.h>
#include "../src/event.h"

char plugin_name[] = "alarm";
char plugin_description[] = "Alarm handling, use the ALLS, ALST and ALRM to list, set or remove an alarm, respectively.";
char plugin_version[] = "1.0";

char *alarm_file = NULL;
struct alarm *alarms = NULL;

/* commands supported by this plugin */
struct cmd_table plugin_commands[] =  {
	{"ALLS", "alarm_list", "Shows all set alarms."},
	{"ALST", "alarm_set", "Set an alarm."},
	{"ALRM", "alarm_remove", "Remove an alarm."},
	{NULL, NULL, NULL}									/* needed for end-of-array checking */
};

struct evt_table plugin_events[] = {
	{"alarm_set" },
	{ NULL }
};

/* plugin initialising */
int plugin_initialise(void) {
	int length;
	char *homedir;

	homedir = getenv("HOME");
	alarms = NULL;

	/* +2 for \0 and a '/' */
	length = strlen(runtimedir) + strlen(DEFAULT_ALARMFILE) + 2;
	alarm_file = (char *)malloc(length * sizeof(char));
	snprintf(alarm_file, length, "%s/%s", runtimedir, DEFAULT_ALARMFILE);

	load_alarms(alarm_file);

	return 0;
}

/* called on plugin unload (by HUP or UNLD) */
int plugin_unload(void) {
	struct alarm *al = alarms, *alm;

	while (al) {
		alm = al;
		al = al->next;

		free(alm->message);
		free(alm);
	}

	oc_writelog("[alarm] alarm plugin unloaded.\n");
}

void save_alarms(const char *alarm_fn) {
	FILE *fp;
	fp = fopen(alarm_fn, "w");
	if (fp) {
		struct alarm *al = alarms;
		while (al) {
			fprintf(fp, "%ld %s\n", al->t_set, al->message);

			al = al->next;
		}

		fclose(fp);
	} else {
		oc_writelog("[alarm] couldn't open %s for writing. Alarms are not saved between sessions.\n", alarm_fn);
	}
}

void load_alarms(const char *alarm_fn) {
	FILE *f;
	f = fopen(alarm_fn, "r");
	if (f) {
		time_t t_set;
		char *message;
		char line[2048];
		struct alarm *al;
		int ptr = 0, length;

		while (fgets(line, 2048, f)) {
			length = strlen(line);
			while (line[ptr] != ' ' && line[ptr++] != '\0');

			if (line[ptr] == ' ') {
				line[length-1] = '\0';
				line[ptr] = '\0';

				t_set = atol(line);
				if (t_set != 0) {
					message = strdup(line+ptr+1);

					al = alarm_create(message);
					al->t_set = t_set;
					al->next = alarms;
					alarms = al;
				}
			}
		}
		fclose(f);
	}

}

struct reply *alarm_list(char **cmdline) {
	struct reply *rpl;
	char *homedir;
	struct alarm *a;

	rpl = malloc (sizeof(struct reply));
	rpl->code = OK;

	rpl->result = buffer_setup();

	a = alarms;
	while (a != NULL) {
		char *ts = create_timestring(a->t_set, "%Y-%m-%d %H:%M:%S");
		rpl->result = buffer_addf(rpl->result, "alarm %d, set at %s: %s\n", a->id, ts, a->message);
		free(ts);
				
		a = a->next;
	}

	return rpl;
}

struct alarm *alarm_create(const char *msg) {
	struct alarm *ret = (struct alarm *)malloc(sizeof(struct alarm));

	if (alarms == NULL) {
		/* we start at 1, so we can use 0 for error handling */
		ret->id = 1;
	} else {
		ret->id = alarms->id + 1;
	}
	ret->message = strdup(msg);
	ret->t_set = time(NULL);
	ret->next = NULL;

	return ret;
}

struct reply *alarm_set(char **cmdline) {
	struct reply *rpl;
	int i = 0;

	rpl = malloc(sizeof(struct reply));
	rpl->result = buffer_setup();

	if (cmdline[1] == NULL) {
		rpl->code = WRN_WRONGPARAM;
		rpl->result = buffer_addf(rpl->result, "You have to supply an alarm message: ALST \"<message>\".\n");
	} else {
		/* alarm message was ok, add to linked list */
		struct alarm *al = alarm_create(cmdline[1]);
		al->next = alarms;
		alarms = al;

		rpl->code = OK;
		rpl->result = buffer_addf(rpl->result, "Alarm set; id: %d, message: %s\n", al->id, al->message);

		save_alarms(alarm_file);

		// and fire our 'alarm_set' event
		event_fire("alarm_set", NULL);
	}

	return rpl;

}

struct reply *alarm_remove(char **cmdline) {
	struct reply *rpl;

	rpl = malloc (sizeof(struct reply));
	rpl->result = buffer_setup();

	if (cmdline[1] == NULL) {
		rpl->code = WRN_WRONGPARAM;
		rpl->result = buffer_addf(rpl->result, "No alarm id given.\n");
	} else {
		int id = atoi(cmdline[1]);
		struct alarm *al = alarms;
		struct alarm *prev = NULL;
		char foundit = 0;

		while (al != NULL) {
			if (al->id == id) {
				if (prev == NULL) {
					alarms = al->next;
				} else {
					prev->next = al->next;
				}
				free(al->message);
				free(al);
				
				/* found it, we're done */
				foundit = 1;
				break;
			} else {
				prev = al;
				al = al->next;
			}
		}


		if (foundit == 0) {
			rpl->code = WRN_WRONGPARAM;
			rpl->result = buffer_addf(rpl->result, "Alarm '%s' is not a valid alarm id.\n", cmdline[1]);
		} else {
			rpl->code = OK;
			rpl->result = buffer_addf(rpl->result, "Alarm %d removed.\n", id);

			save_alarms(alarm_file);
		}

	}

	return rpl;
}
